/*
 * HttSquash, http server scanner, banner grabber and data retriever
 *
 * Copyright (C) 2009  Acri Emanuele <crossbower@gmail.com>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "httsquash.h"

/*
 * Write in memory (instead of file).
 * Taken from:
 * http://curl.haxx.se/lxr/source/docs/examples/getinmemory.c
 */
size_t WriteMemoryCallback(void *ptr, size_t size, size_t nmemb, void *data)
{
    size_t realsize = size * nmemb;

    struct MemoryStruct *mem = (struct MemoryStruct *)data;

    mem->memory = (char *) realloc(mem->memory, mem->size + realsize + 1);

    if (mem->memory) {
        memcpy(&(mem->memory[mem->size]), ptr, realsize);
        mem->size += realsize;
        mem->memory[mem->size] = 0;
    }

    return realsize;
}

/*
 * Function called by scanloop()
 * Executed as a thread
 * Gets the ip address and retrieve informations from it
 */
void *scanstep(void *arg)
{
    CURL *curl;
    CURLcode res;
    struct MemoryStruct header, data;

    char addr[MIN_ARG];
    char buff[MAX_DATA];

    struct thread_data *th_data;

    // clear structures
    memset(&header, 0, sizeof(header));
    memset(&data, 0, sizeof(data));

    // cast arg
    th_data = (struct thread_data *) arg;

    // current address
    if(th_data->ipr->ipver==VER_IPV4)
        sprintf(addr, "%s.%u", th_data->ipr->block, th_data->curr);
    else
        sprintf(addr, "%s:%x", th_data->ipr->block, th_data->curr);

    // initialize curl
    curl = curl_easy_init();
    if (curl == NULL) {
        fprintf(stderr, "Error: initialize libcurl\n");
        pthread_exit(NULL);
    }

    // debug
    //curl_easy_setopt(curl, CURLOPT_VERBOSE, 1);

    // set port and timeout
    curl_easy_setopt(curl, CURLOPT_PORT, th_data->ipr->port);
    curl_easy_setopt(curl, CURLOPT_TIMEOUT, th_data->ipr->time);


    // set request type
    curl_easy_setopt(curl, CURLOPT_CUSTOMREQUEST, th_data->ipr->type);

    // don't verify SSL certificate
    curl_easy_setopt(curl, CURLOPT_SSL_VERIFYPEER, 0);
    curl_easy_setopt(curl, CURLOPT_SSL_VERIFYHOST, 0);

    // set request url
    snprintf(buff, MAX_DATA-1, "%s%s%s", th_data->ipr->https ? "https://" : "http://", addr, th_data->ipr->url);
    curl_easy_setopt(curl, CURLOPT_URL, buff);


    // set personalized headers
    if(th_data->ipr->headers) {
        curl_easy_setopt(curl, CURLOPT_HTTPHEADER, th_data->ipr->headers);
    }


    // set memory write callback
    curl_easy_setopt(curl, CURLOPT_WRITEFUNCTION, WriteMemoryCallback);
    curl_easy_setopt(curl, CURLOPT_WRITEHEADER, (void *)&header);
    curl_easy_setopt(curl, CURLOPT_WRITEDATA, (void *)&data);

    // get result
    res = curl_easy_perform(curl);
    if(res!=0) {
        curl_easy_cleanup(curl);
        pthread_exit(NULL);
    }

    // *** Print All ***
    printf("FOUND: %s %d\n%s%s%s%s%s",

            // header
            addr,
            th_data->ipr->port,
            header.memory,

            // data
            th_data->ipr->body ? "DATA:\n" : "",
            th_data->ipr->body ? data.memory : "",
            th_data->ipr->body ? "\n\n" : "",

            // cookie jar
            th_data->ipr->cookiejar ? "%%\n\n" : "");

    // free structures
    if(header.memory) free(header.memory);
    if(data.memory) free(data.memory);

    // cleanup
    curl_easy_cleanup(curl);

    // thread exit
    pthread_exit(NULL);
}

/*
 * Scan loop that assembles current ip address
 * and exec a callback function
 */
void scanloop(struct iprange *ipr, int max_th)
{
    int net_end, curr, curr2;
    int th_count=0;

    // array of threads structures
    pthread_t *thread;
    struct thread_data th_data[max_th];
    pthread_attr_t attr;

    curr = ipr->start;

    // alloc and set attr
    thread = (pthread_t *) malloc(sizeof(pthread_t *) * (max_th+1));
    pthread_attr_init(&attr);
    //pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_JOINABLE);

    // no extended range
    if(ipr->range_ext==0) {
        while(curr <= ipr->end) {

            th_data[th_count].ipr  = ipr;
            th_data[th_count].curr = curr;

            // create new thread
            pthread_create(&thread[th_count], &attr, scanstep, &th_data[th_count]);


            // increment counters
            curr++;

            // check threads number
            if(th_count >= ipr->max_threads) {

                int k=0;

                // wait for the other threads
                for(; th_count; th_count--, k++) {
                    pthread_join(thread[th_count], NULL);
                }

            } else {
                th_count++;
            }
        }
    }

    // extended range (only ipv4)
    else {
        curr2 = ipr->start2;

        while(curr2 <= ipr->end2) {

            // set net_end
            if(curr2==ipr->end2) {
                net_end=ipr->end;
            }
            else {
                net_end=254;
            }

            while(curr <= net_end) {

                th_data[th_count].ipr = ipr;
                th_data[th_count].curr = curr;

                // create new thread
                pthread_create(&thread[th_count], &attr, scanstep, &th_data[th_count]);

                // increment counters
                curr++;

                // check threads number
                if(th_count >= ipr->max_threads) {

                    int k=0;

                    // wait for the other threads
                    for(; th_count; th_count--, k++) {
                        pthread_kill(thread[k], SIGTERM);
                    }

                } else {
                    th_count++;
                }
            }

            curr = 0;
            curr2++;
        }
    }

    // check threads number
    if(th_count) {

        int k=0;

        // wait for the other threads
        for(; th_count; th_count--, k++) {
            pthread_join(thread[k], NULL);
        }
    }

    free(thread);

    return;
}

/*
 * Initialize the program
 */
struct iprange *prog_initialize()
{
    // alloc ipr
    struct iprange *ipr = (struct iprange *) malloc(sizeof(struct iprange));
    memset(ipr, 0, sizeof(struct iprange));

    return ipr;
}

/*
 * Cleanup the program
 */
void prog_cleanup(struct iprange *ipr)
{
    // free header list
    if(ipr->headers) curl_slist_free_all(ipr->headers);

    // dealloc ipr
    free(ipr);
}

/*
 * Main function
 */
int main(int argc, char **argv)
{
    struct iprange *ipr = NULL;

    // Initialize the program
    ipr = prog_initialize();

    // Parse options
    parseopt(ipr, argc, argv);

    // Init scan loop
    scanloop(ipr, ipr->max_threads);

    // Cleanup the program
    prog_cleanup(ipr);

    // exit main thread
    pthread_exit(NULL);
}

